// vacsearch.cpp : Defines the entry point for the console application.
#pragma warning(disable:4786) 
// The debugger can't handle symbols more than 255 characters long.
// STL often creates symbols longer than that, so we disable the warning by this "pragma" line.
#include <iostream> //Use this instead of iostream.h for standard C++
#include <vector> //STL template.  Note that you don't have .h at end.
#include <deque> //STL template, like a vector with push_front.  Pronounced like "deck"
using namespace std; //Need this line to use STL templates with Visual C++.

#define DEPTHFIRST
/* Comment out the DEPTHFIRST to do a Breadth first search. */
//===========================Position====================
class Position
{
public:
	int _col;
	int _row;
	Position():_col(0),_row(0){}
	Position(int col, int row):_col(col), _row(row){}
	Position(const Position &pos):_col(pos._col),_row(pos._row){}
	bool home(){return _col==0 && _row == 0;}
	friend ostream& operator<<( ostream& os, Position &pos);
	friend operator==(Position p, Position q){return p._col==q._col && p._row==q._row;}
};

ostream& operator<< ( ostream& os, Position &pos)
{
	static char letter[26] = {'A','B','C','D','E','F','G','H','I','J','K','L','M','N','O','P','Q','R','S','T','U','V','W','X','Y','Z'};
	os << letter[pos._col % 26] << pos._row + 1; //Reuse col letters if more than 26 cols.
	return os;
}
//===========================World====================
class World
{
private:
	int _colsize;
	int _rowsize;
	vector<Position> _dirtposition;
public:
	World(){}
	World(int colsize, int rowsize):_colsize(colsize), _rowsize(rowsize){}
	int rowsize(){return _rowsize;}
	int colsize(){return _colsize;}
	void addDirt(Position pos){_dirtposition.push_back(pos);}
	void addDirt(int poscode){addDirt(Position(poscode%colsize(), poscode/colsize()));}
	bool hasDirt(const Position &pos);
	friend ostream& operator<<( ostream& os, World &pos);
};

ostream& operator<<( ostream& os, World &world)
{
	os << world._colsize << " columns. " << world._rowsize << " rows. Dirt in positions: ";
	for (int i=0; i<world._dirtposition.size(); i++)
		os << world._dirtposition[i] << " ";
	os << ".\n";
	return os;
}	

bool World::hasDirt(const Position &pos)
{
	for (int i=0; i<_dirtposition.size(); i++)
		if (pos == _dirtposition[i])
			return true;
	return false;
}	

//===========================State====================
class State
{
public:
	Position _pos;
	int _dir;
	vector<Position> _dirtposition;
	bool _hitwall;
	bool isLegal(){return !_hitwall;}
	int dirtscore(){return _dirtposition.size();}
	State():_dir(0), _hitwall(false){} /* Default Position constructor sets (0,0); default _dirtposition is
		empty list. */
	State(const Position &pos, int dir, const vector<Position> &dirtposition, bool hitwall):
		_pos(pos), _dir(dir), _dirtposition(dirtposition), _hitwall(hitwall){}
	bool hasDirt(const Position &pos);
	State turnleft(World *pworld){return State(_pos, (_dir+1)%4, _dirtposition, false);}
	State turnright(World *pworld){return State(_pos, (_dir+3)%4, _dirtposition, false);}
	State goforward(World *pworld);
	bool attemptsuck(World *pworld);
	friend ostream& operator<<( ostream& os, State &state);
	friend bool operator==(const State &s1, const State &s2){return s1._pos == s2._pos && s1._dir==s2._dir &&
		s1._hitwall == s2._hitwall && s1._dirtposition == s2._dirtposition;}
};

ostream& operator<< ( ostream& os, State &state)
{
	static char dirletter[4] = {'e','n','w','s'};
	os << state._pos << " " << dirletter[state._dir % 4] << " " << state.dirtscore();
	return os;
}

bool State::hasDirt(const Position &pos)
{
	for (int i=0; i<_dirtposition.size(); i++)
		if (pos == _dirtposition[i])
			return true;
	return false;
}	

State State::goforward(World *pworld)
{
/* We think of the origin of the world as (0,0) at upper left corner, and we view _col as increasing
left to right and _row as increasing top to bottom. */
	int newcol = _pos._col;
	int newrow = _pos._row;
	bool hitwall = false;
	switch(_dir)
	{
		case 0: //East
			newcol++;
			if (newcol >= pworld->colsize())
				{newcol--; hitwall = true;}
			break;
		case 1: //North
			newrow--;
			if (newrow < 0)
				{newrow++; hitwall = true;}
			break;
		case 2: //West
			newcol--;
			if (newcol < 0)
				{newcol++; hitwall = true;}
			break;
		case 3: //South
			newrow++;
			if (newrow >= pworld->rowsize())
				{newrow--; hitwall = true;}
			break;
	}
	Position newpos(newcol, newrow);
	State newstate(newpos, _dir, _dirtposition, hitwall);
	newstate.attemptsuck(pworld);
	return newstate;
}

bool State::attemptsuck(World *pworld)
{
	if (pworld->hasDirt(_pos) && 	 //If the new position has a piece of dirt
		!hasDirt(_pos)) //And if newpos is not already in the newstates dirtposition vector
	{
		_dirtposition.push_back(_pos);
		return true;
	}
	return false;
}

//============================StateSet=====================

class StateSet : public deque<State>
{
public:
	bool find(const State &s);
};

bool StateSet::find(const State &s)
{
	for (int i=0; i<size(); i++)
		if (s == at(i))
			return true;
	return false;
}
//============================Node=====================

class Node 
{
private:
	Node * _pparent;
	State _state;
	int _depth;
	int _childcount;
public:
	static int MAXDIRTCOUNT;
	static int NODESONHEAP;
	static int NODESMEMORYREQ;
	Node():_pparent(NULL), _depth(1), _childcount(0){NODESONHEAP++;} //Calls default State constructor.
	Node(State &state, Node *pnode = NULL);
	~Node();
	StateSet priorpathstates(); // Use this to avoid repeated states in Node::expand.
	vector<Node*> expand(int maxdepth, World *pworld);
	bool isGoal(){return _state._pos.home() && _state.dirtscore() == MAXDIRTCOUNT;}
	bool isParent(){return _childcount > 0;}
	friend ostream& operator<<( ostream& os, Node *pnode);
};

int Node::MAXDIRTCOUNT = 0;
int Node::NODESONHEAP = 0;
int Node::NODESMEMORYREQ = 0;

Node::Node(State &state, Node *pnode):
	_pparent(pnode),
	_state(state),
	_childcount(0)
{
	if (!_pparent)
		_depth=1;
	else
	{
		_depth = _pparent->_depth+1;
		_pparent->_childcount++;
	}
	NODESONHEAP++;
	/* I want to track the "high water mark" of the maximum nodes allocated in memory at
		one time so that I can talk about the "space complexity" of my search algorithms. */
	if (NODESONHEAP > NODESMEMORYREQ)
		NODESMEMORYREQ = NODESONHEAP;
}

Node::~Node()
{
	if (_pparent)
	{
		(_pparent->_childcount)--;
		if (!(_pparent->_childcount))
			delete _pparent;
	}
	NODESONHEAP--;
}


ostream& operator<< ( ostream& os, Node *pnode )
{
	if (!pnode)
		return os;
	StateSet priorstates = pnode->priorpathstates(); 
	for (int i=0; i< priorstates.size(); i++)
		os << priorstates[i] << ", ";
	os << pnode->_state <<".\n";
	return os;
}

StateSet Node::priorpathstates()
{ 
	StateSet statedeque;
	Node* pcur = _pparent;
	while (pcur != NULL)
	{
		statedeque.push_front(pcur->_state);
		pcur = pcur->_pparent;
	}
	return statedeque;
}

vector<Node*> Node::expand(int maxdepth, World *pworld)
{
	vector<Node*> newnodes; //Default array is empty.
	if (_depth>=maxdepth)
		return newnodes; 
	StateSet oldpathstates = priorpathstates(); //States already visited.
	State newstate;
//(1)Your code????????????????????????????????????????????????????????????	
	return newnodes;
}

//============================SearchQueue=====================

class SearchQueue : public deque<Node*>
{
public:
	static int NODESHANDLED;

	SearchQueue(World *pworld);
	~SearchQueue();
	void addToQueue(vector<Node*> &newnodes);
};

int SearchQueue::NODESHANDLED = 0;

SearchQueue::SearchQueue(World *pworld)
{
	State defaultstate;
	defaultstate.attemptsuck(pworld);
	Node * pnode = new Node(defaultstate);
	push_back(pnode);
	NODESHANDLED = 1;
}

SearchQueue::~SearchQueue()
{
	while (!empty())
	{
		Node* ptestnode = front();	
		pop_front();	//Remove head node.
		delete ptestnode;
	}
}

void SearchQueue::addToQueue(vector<Node*> &newnodes)
{
	//(2) Your code?????????????????????????????????????????????????????
	NODESHANDLED += newnodes.size();
}

//============================main=====================

int main(int argc, char* argv[])
{
	cout <<
"Your Name's Feb, ??, 2002, Vacsearch Program, simulating a robot vacuum cleaner.\n";
	cout <<
"Blind search to find short path in a grid world, hitting some spots with dirt.\n";
	cout <<
"Three or more integer command line parameters are requred.\n";
	cout <<
"vacsearch cols rows dirtpos1 dirtpos2 dirtpos2 ... \n";
	cout <<
"If you don't give any args, this default is used:  vacsearch 3 3 0 1 \n";
	cout <<
"First two params specify the width and height of the world.\n";
	cout <<
"The additional dirtpos params code up positions as numbers.\n";
	cout <<
"We think of positions as labelled from left to right and top to bottom,\n";
	cout <<
"starting with (0,0) at upper left, which has dirtpos code 0.\n";
	cout <<
"(The dirtpos params are of form x+y*cols, where cols is the first parameter.)\n\n\n";
	World *pworld;
	if (argc >= 4) //The first, or 0th arg is the name of the executable.
	{
		pworld = new World(atoi(argv[1]), atoi(argv[2]));
		Node::MAXDIRTCOUNT = 0;
		for (int i=3; i<argc; i++)
		{
			pworld->addDirt(atoi(argv[i])); 
			Node::MAXDIRTCOUNT++;
		}
	}
	else //Not enough args
	{
		pworld = new World(3, 3);
		pworld->addDirt(0); 
		pworld->addDirt(4); 
		Node::MAXDIRTCOUNT = 2;
	}
	cout << "Now searching for optimal path in the following world.\n\n" << *pworld << "\n";
	cout << "Search depth: "; 
	SearchQueue *pnode_queue = NULL;
	bool solutionfound = false;
	int maxdepth = 0;
	Node *ptestnode = NULL;
//(3) Your code ??????????????????????????????????????????????????
	cout << "\n\nDone searching for optimal path in the following world.\n\n" <<
		*pworld << "\n" <<
		"Found this sequence for the path, where states are listed as \n" <<
		"Position(LetterNumber), Direction(e,n,w,s), Dirt Sucked.\n\n" << ptestnode <<
		"\nSearch was iterative deepening and " << 
#ifdef DEPTHFIRST
		"DEPTH FIRST.\n" <<
#else //not DEPTHFIRST
		"BREADTH FIRST.\n" <<
#endif //DEPTHFIRST
		"\nTime complexity: Searched " <<	SearchQueue::NODESHANDLED << " nodes.\n";
		cout << "Space complexity: Max nodes used at one time was "<<	Node::NODESMEMORYREQ << ".\n";
		delete ptestnode;
		delete pnode_queue;
		cout <<	"Garbage collection: " << Node::NODESONHEAP << " unreleased nodes still on heap.\n";
	cout <<"\n\nYour Name's Feb ??, 2002, Vacsearch Program, simulating a robot vacuum cleaner.\n";
	cout <<"Press ENTER key to exit.";
	getchar(); //Wait for a keypress.
	delete pworld;
	return 0;
}
